################################################################################
#                                                                              #
# Convective Mode Prediction System (CoMPS)                                    #
# Storm Mode Variation                                                         #
# Version 1.1.1                                                                #
#                                                                              #
# Please see "README.txt" file before using                                    #
#                                                                              #
# This Python code is provided as is. Modification of the code is permissible, #
# but you may not distribute this code or modified versions of this code       #
# without clearly crediting:                                                   #
#                                                                              #
# David Stang                                                                  #
# (C) 2021                                                                     #
#                                                                              #
# If modifications are made, you must clearly indicate who made the            #
# modifications and what sections of the file were modified. If distributing   #
# a modified version, you must also include an unmodified version.             #
#                                                                              #
# For complete background information about this model, please read the        #
# following publication:                                                       #
# http://davidstang.org/research/master.pdf                                    #
#                                                                              #
# Any questions regarding this model should be sent to:                        #
# david.w.stang-1@ou.edu                                                       #
#                                                                              #
# Model outputs should NOT be used to make immediate life-saving decisions.    #
# All outputs are provided as is. By using this program and its output, you    #
# voluntarily assume all risks associated with its operation.                  #
#                                                                              #
################################################################################


print("""
Convective Mode Prediction System (CoMPS)
Storm Mode Variation
Version 1.1.1

(C) 2021 by David Stang

If you use this model or its relevant concepts to prepare a publication, please cite:
Stang, D., 2021: A Statistical Approach To Diagnosing Storm Mode. M.S. Thesis, Department of Atmospheric and Geographic Sciences, The University of Oklahoma (Norman Campus), 47 pp., https://shareok.org/handle/11244/332302
""")

import requests
import os
import io
import pygrib
import numpy as np
import matplotlib.pyplot as plt
import math
import sys
import cartopy.crs as ccrs              # Cartopy was used since Basemap is deprecated.
import cartopy.feature as cf
from matplotlib.colors import *

# Custom object that easily stores model data in a three-dimensional dataset.
class Field():
    def __init__(self, x, fieldName, modelName, FH):
        self.name = fieldName
        self.model = modelName
        self.forecastHour = FH

        self.data = []
        self.latitude = []
        self.longitude = []
        self.level = []

        for i in range(0, len(x)):
            message = x[i]

            if (message.level > 1000 or message.level < 100):
                continue

            if (len(self.level) == 0):
                self.latitude, self.longitude = message.latlons()

            if (message["parameterName"] == fieldName):
                self.data = [message.data()[0]] + self.data
                self.level = [message.level] + self.level


    # Function to extract information out of a column. The returned value is a list,
    # ordered from lowest elevation to highest elevation.

    def getProfile(self, i, j):
        result = []
        
        for z in range(0, len(self.level)):
            result.append(self.data[z][i, j])
            
        return result
    

    # Function to extract the highest entry out of a column.

    def getMax(self, i, j):
        result = self.data[0][i, j]
        
        for z in range(0, len(self.level)):
            if (self.data[z][i, j] > result):
                result = self.data[z][i, j]
                
        return result

    
    # This function is immensely helpful for conserving memory. We don't need the entire
    # grid of values produced by the global models. So, we'll crop out a subsection that
    # we are interested in.

    def export(self):
        parentDirectory = self.model + "/FH" + fitLength(self.forecastHour, 2) + "/" + self.name + "/"
        
        # Create this path if it doesn't already exist.
        if (not(os.path.exists(parentDirectory))):
            if (not(os.path.exists(self.model))):
                os.mkdir(self.model)
            if (not(os.path.exists(self.model + "/FH" + fitLength(self.forecastHour, 2)))):
                os.mkdir(self.model + "/FH" + fitLength(self.forecastHour, 2))
            os.mkdir(parentDirectory)
        
        makeSpreadsheet(self.latitude, parentDirectory+"LATITUDE.csv")
        makeSpreadsheet(self.longitude, parentDirectory+"LONGITUDE.csv")
        
        for z in range(0, len(self.level)):
            makeSpreadsheet(self.data[z], parentDirectory + str(self.level[z]) + ".csv")
            

    # We don't need the entire grid from the GFS, so here's a function to crop out the
    # unwanted data.

    def crop(self, latBound, lonBound):
        try:
            self.latitude  =  self.latitude[latBound[0]:latBound[1], lonBound[0]:lonBound[1]]
            self.longitude = self.longitude[latBound[0]:latBound[1], lonBound[0]:lonBound[1]]
        
            for z in range(0, len(self.level)):
                self.data[z] = self.data[z][latBound[0]:latBound[1], lonBound[0]:lonBound[1]]
        except:
            for z in range(0, len(self.level)):
                self.data[z] = self.data[z][latBound[0]:latBound[1], lonBound[0]:lonBound[1]]


    # Get the standard deviation of the values in a column.

    def getStandardDeviation(self, i, j):
        result = 0
        theSum = 0
        n      = 0

        for z in range(0, len(self.level)):
            theSum = theSum + self.data[z][i, j]
            n = n + 1

        theMean = theSum / n

        for z in range(0, len(self.level)):
            result = result + (self.data[z][i, j] - theMean)**2

        return math.sqrt(result / n)

        
# Some numerical arguments need to be a certain number of digits, this method is an easy
# way to ensure this happens. Ex: fitLength(2, 4) would return "0002" instead of "2".

def fitLength(x, theLength):
    asString = str(x)
    while (len(asString) < theLength):
        asString = "0" + asString
    
    return asString


# Function for converting latitude from mathematical convention to standard geographic
# convention.

def formatLatitude(x):
    y = str(x)

    if (x > 0):
        y = y + " N"
    if (x < 0):
        y = y + " S"

    deficit = max(7 - len(y), 0)

    y = (" " * deficit) + y

    return y


# Function for converting longitude from mathematical convention to standard geographic
# convention.

def formatLongitude(x):
    y = str(x)

    if (x > 0):
        y = y + " E"
    if (x < 0):
        y = y + " W"

    deficit = max(7 - len(y), 0)

    y = (" " * deficit) + y

    return y


# Method for retrieving data from the world wide web.

def loadResource(theURL):
    rawData = requests.get(theURL)
    return rawData.content


# Method used for macroscopic diagnosis of well-defined boundaries.

def weightedLinearRegression(pointList, weightList, n):
    if (len(pointList) <= 1 or len(weightList) <= 1):
        return -999, -999
    
    sigmaWXX = 0
    sigmaWXY = 0
    sigmaWX  = 0
    sigmaWY  = 0
    sigmaW   = 0
    sigmaX   = 0
    
    for i in range(0, len(pointList)):
        sigmaWXX = sigmaWXX + weightList[i]**n * pointList[i][0] * pointList[i][0]
        sigmaWXY = sigmaWXY + weightList[i]**n * pointList[i][0] * pointList[i][1]
        sigmaWX  = sigmaWX  + weightList[i]**n * pointList[i][0]
        sigmaWY  = sigmaWY  + weightList[i]**n * pointList[i][1]
        sigmaW   = sigmaW   + weightList[i]**n
        sigmaX   = sigmaX   + pointList[i][0]
        
    try:
        a = ( sigmaWXY - ((sigmaX * sigmaWY) / sigmaW)) / ( sigmaWXX - ((sigmaX * sigmaWX) / sigmaW) )
    except:
        a = -999
        
    try:
        b = (sigmaWY - a * sigmaWX) / sigmaW
    except:
        b = -999
        
    return a, b


# Simple function to calculate mean of a list

def mean(x):
    return (sum(x) / len(x))


# Function to calculate a vectorial mean of the wind field.

def meanWind(rawSounding, u, v):
    soundingT = -9999
    soundingD = -9999
    z = 0

    while (soundingT <= -999 or soundingD <= -999):
        soundingT = float( rawSounding[z][2] )
        soundingD = float( rawSounding[z][3] )

        z = z + 1

    SB_LCL = 1000 * ((soundingT - soundingD) / 8.0)

    uList = []
    vList = []

    EL_strikes     = 0
    EL_max_strikes = 3

    lastT = -999

    for z in range(0, len(rawSounding)):
        if ( rawSounding[z][1] >= SB_LCL ):
            uList.append( u[z] )
            vList.append( v[z] )

        if ( rawSounding[z][2] > lastT and rawSounding[z][1] >= SB_LCL ):
            EL_strikes = EL_strikes + ( (rawSounding[z][2] + 273.15) / (lastT + 273.15) )
            if (EL_strikes >= EL_max_strikes):
                break
        else:
            EL_strikes = 0

        lastT = rawSounding[z][2]

    uMean = mean(uList)
    vMean = mean(vList)

    meanAngle = math.atan2(vMean, uMean) * (180 / math.pi)
    meanSpeed = math.sqrt(uMean**2 + vMean**2)

    return meanSpeed, meanAngle


# Method for saving the forecast grid to a compact binary file. The saved file can
# be loaded using the program "CFG.py" in the folder "cfg-file".

def makeDataset(longitude, latitude, x, fileName):
    y = b''

    rCount = len(x[:, 0])
    cCount = len(x[0, :])

    # Header
    y = y + b'CFG'

    # Version number
    y = y + b'\x01\x00\x00'

    # Latitude divisor/multiplier (10**2)
    y = y + b'\x02'

    # Longitude divisor/multiplier (10**2)
    y = y + b'\x02'

    # Data Divisor/multiplier (10**2)
    y = y + b'\x02'

    # Longitude size (bytes)
    y = y + b'\x02'

    # Latitude size (bytes)
    y = y + b'\x02'

    # Entry size (bytes)
    y = y + b'\x01'

    # Row count
    y = y + rCount.to_bytes(2, "big")

    # Column count
    y = y + cCount.to_bytes(2, "big")

    for i in range(0, len(latitude)):
        value = int(100 * round(latitude[i], 2))

        y = y + value.to_bytes(2, "big")

    for i in range(0, len(longitude)):
        value = int(100 * round(longitude[i], 2))

        y = y + value.to_bytes(2, "big")

    for i in range(0, len(x[:, 0])):
        for j in range(0, len(x[i, :])):
            value = 0xff

            if (not(x.mask[i, j])):
                value = int(100 * round(x[i, j], 2))

            y = y + value.to_bytes(1, "big")

    pointer = open(fileName, 'wb')
    pointer.write(y)
    pointer.close()


# A method for exporting a two-dimensional array into a spreadsheet for quick
# and easy viewing (unused in the original version)

def makeSpreadsheet(x, fileName):
    pointer = open(fileName, 'w')
    
    OUTPUT = ""
    
    for i in range(0, len(x)):
        for j in range(0, len(x[i])):
            OUTPUT = OUTPUT + str(round(x[i, j], 2)) + ","
            
        OUTPUT = OUTPUT[0:-1] + "\n"
        
    pointer.write(OUTPUT[0:-1])
    pointer.close()

    
# Calculate dewpoint given relative humidity and air temperature.

def dewpoint(RH, T):
    RV = 461.5
    LV = 2.5e06
    
    D = 1 / ((1 / T) - (RV / LV) * math.log(max(RH, 0.01)))
    
    return D


# Calculate the mean profile in a sounding (vertical sample).

def soundingMean(x):
    y = []

    x = np.array(x)

    x = x.mean(0)

    return x


# This function takes in a list of pressures, heights, temperatures, and relative
# humidities. The returned value is a list that can easily be plotted on a sounding.
# Note that we calculate the dewpoint list from the list of relative humidities (which
# are reported as a percentage and not a decimal).

def soundingGrid(P, z, T, R, u, v, omega, n=0):
    D = []
    
    for i in range(0, len(P)):
        D.append(dewpoint(0.01 * R[i], T[i]))        
        
    y = []
    
    for i in range(0, len(P)):
        y.append([P[i], z[i], T[i], D[i], u[i], v[i], omega[i]])
    
    return y


# Method for calculating a Gaussian (normal) probability.

def gaussianProbability(offset, sigma, start_x, end_x, delta_x, coefficient = 2):
    x = start_x
    area = 0
    baseArea = sigma * math.sqrt(math.pi * coefficient)
    
    while x < end_x:
        area = area + delta_x * math.exp(-1 * ((x - offset)**2) / (coefficient * sigma**2))
        x = x + delta_x
        
    return (area / baseArea)


# Method for performing linear interpolations.

def interpolate(x, y, z, reverse=False):
    bestUpper = -1
    bestLower = -1
    theValue  = 0
        
    if (reverse):
        lastValue = -1

        for h in range(1, len(x)):
            if (y[h] <= z and x[h] > -999 and x[h-1] > -999):
                bestUpper = h
                lastValue = h
                break
            
        if (bestUpper < 0):
            bestUpper = lastValue
    else:
        lastValue = -1
        
        for h in range(1, len(x)):
            if (y[h] >= z and x[h] > -999 and x[h-1] > -999):
                bestUpper = h
                lastValue = h
                break
            
        if (bestUpper < 0):
            bestUpper = lastValue
                
    if (y[bestUpper] == z):
        return x[bestUpper]
    
    for h in range(bestUpper-1, -1, -1):
        if (y[h] > -999):
            bestLower = h
            break
                
    if (bestUpper == bestLower):
        theValue = x[bestUpper]
    else:
        dx_dy = (x[bestUpper] - x[bestLower]) / (y[bestUpper] - y[bestLower])
        
        theValue = dx_dy * (z - y[bestLower]) + x[bestLower]
        
    return theValue


# Function to calculate virtual temperature given air temperature and dewpoint. NOTE: The
# reason for the unit arguments is because I was testing this function on multiple
# different observations that were all in different units. All the temperatures in the
# models are reported in Kelvin.

def virtualTemperature(T, D, unitT, unitD):
    T = float(T)
    D = float(D)
    
    if (unitT == "C"):
        T = T + 273.15
    elif (unitT == "F"):
        T = (T - 32) * (5.0 / 9.0) + 273.15
    elif (unitT == "K"):
        T = T
    else:
        raise ValueError("Unsupported unit for Temperature")
        
    if (unitT == "C"):
        D = D + 273.15
    elif (unitT == "F"):
        D = (D - 32) * (5.0 / 9.0) + 273.15
    elif (unitT == "K"):
        D = D
    else:
        raise ValueError("Unsupported unit for Dewpoint")
        
    LV = 2.5e06
    RV = 416.5
    
    RH = 100 * math.exp( (LV/RV) * ( (1/T) - (1/D)))
    eS = 611 * math.exp( (LV/RV) * ( (1/273) - (1/T)))
    
    w = 0.622 * (RH / 100) * eS / 100000
    
    Tv = T * (1 + 0.61 * w)
    
    return Tv


# Function to calculate moist adiabatic lapse rate given temperature and pressure. This
# was also tested using a variety of different units (hence the extra parameters). The
# returned value is a lapse rate in K / km. SIDE NOTE: The equation is a pain, but not
# nearly as painful as doing it by hand...

def moistLapseRate(T, P, unitT, unitP):
    T = float(T)
    P = float(P)
    
    if (unitT == "C"):
        T = T + 273.15
    elif (unitT == "F"):
        T = (T - 32) * (5.0 / 9.0) + 273.15
    elif (unitT == "K"):
        T = T
    else:
        raise ValueError("Unsupported unit for Temperature")
    
    
    if (unitP == "hPa" or unitP == "mb"):
        P = P * 100
    elif (unitP == "kPa"):
        P = P * 1000
    elif (unitP == "Pa"):
        P = P
    else:
        raise ValueError("Unsupported unit for Pressure")
        
    LV = float(2.5e06)
    RV = 461.5
    eS = 611.0 * math.exp( (LV/RV) * ( (1/273) - (1/T)))
    wS = (0.622 * eS) / P
    
    LR = 9.8 * ((1 + (wS * LV) / (287 * T)) / (1 + (LV*LV*wS * 0.622) / (1004.67 * 287.0 * T*T)))

    return LR


# Given temperature and pressure, calculate mixing ratio. Returned value is g/g (or
# kg/kg).

def mixingRatio(T, P):
    T = float(T)
    P = float(P)
    
    LV = 2.5e06
    RV = 461.5

    try:
        eS = 611 * math.exp( (LV/RV) * ( (1/273.15) - (1/T)))
    except:
        print('\aCould not calculate mixing ratio: T = ' + str(T) + ', ' + str(P))

    wS = (0.622 * eS) / P
    
    return wS


# Clausius-Clapeyron equation (the derived version; not the empirical version).

def e_s(T):
    x = 611 * math.exp( 5423 * ( (1/273) - (1/T) ) )
    return x


# Function to calculate Theta-E (in K) given temperature, relative humidity, and
# pressure.

def thetaE(T, RH, P):
    T = float(T)
    RH= float(RH)
    w = (0.622 * (RH / 100.0) * e_s(T)) / P
    x = T * math.exp( (2500000.0 * w) / (1004.67 * T))
    return x


# Function to calculate a two-dimensional array of Theta-E given arrays of temperature,
# relative humidity, and pressure.

def thetaGrid(T, RH, P):
    y = np.zeros([len(T[:,0]), len(T[0,:])])
    
    for i in range(0, len(y[:,0])):
        for j in range(0, len(y[0, :])):
            y[i, j] = thetaE(T[i, j], RH[i, j], P)
            
    return y


# Function to estimate front angle given a grid of Theta-E values. NOTE: The 'size'
# argument determines how many grid points we'll examine at a time. The default value is
# 8, but the actual research used a value of 16, which works out to subsections that are
# roughly 1000 km by 1000 km (on the order of synoptic-scale).

def frontAngleGrid(thetaArray, size=8):
    offsetX = (len(thetaArray[0, :]) % size) / 2.0
    offsetY = (len(thetaArray[:, 0]) % size) / 2.0

    frontAngle = np.zeros([len(thetaArray[:, 0]), len(thetaArray[0, :])])
    gradTheta  = np.zeros([len(thetaArray[:, 0]), len(thetaArray[0, :])])
    gradGrid   = np.zeros([len(thetaArray[:, 0]), len(thetaArray[0, :])])

    if (int(offsetX) != offsetX):
        print('\aWARNING: x-dimension could not be centered')

    if (int(offsetY) != offsetY):
        print('\aWARNING: y-dimension could not be centered')

    offsetX = int(offsetX)
    offsetY = int(offsetY)

    maxY = len(thetaArray[:,0])
    maxX = len(thetaArray[0,:])

    for y in range(offsetY, len(thetaArray[:, 0]) - offsetY, size):
        for x in range(offsetX, len(thetaArray[0, :]) - offsetX, size):
            subset = thetaArray[y:y+size,x:x+size]    # Grab a subsection.

            gradPoint = []
            gradWeight= []

            grad   = np.zeros([size, size])

            for sY in range(0, len(subset[:, 0])):
                for sX in range(0, len(subset[0, :])):
                    # Check to make sure we stay within the grid.
                    if (y+sY+1 >= maxY or y+sY-1 < 0 or x+sX+1 >= maxX or x+sX-1 < 0):
                        continue

                    # Calculate the magnitude of the Theta-E gradient.
                    dT_dx = thetaArray[y+sY, x+sX+1] - thetaArray[y+sY, x+sX-1]
                    dT_dy = thetaArray[y+sY-1, x+sX] - thetaArray[y+sY+1, x+sX]

                    gradT = math.sqrt((dT_dy**2) + (dT_dx**2))

                    grad[sY, sX] = gradT
                    gradTheta[y+sY, x+sX] = gradT

            # Look for strong gradients initially. If we can't find strong gradients,
            # look for weaker gradients (but not gradients that are too weak).
            minGrad = 50          # This works out to roughly 0.8 K / km. The number in
                                  # below loop (10) works out to roughly 0.2 K / km.

            while (len(gradPoint) < 2 and minGrad >= 10.0):
                for sY in range(1, len(subset[:, 0])-1):
                    for sX in range(1, len(subset[0, :])-1):
                        if (grad[sY, sX] > grad[sY+1, sX] and grad[sY, sX] > grad[sY-1, sX] and grad[sY, sX] > grad[sY, sX-1] and grad[sY, sX] > grad[sY, sX+1] and grad[sY, sX] > grad[sY+1, sX+1] and grad[sY, sX] > grad[sY-1, sX+1] and grad[sY, sX] > grad[sY+1, sX-1] and grad[sY, sX] > grad[sY-1, sX+1]):

                            # Look for local maxima of the Theta-E gradient, but also
                            # make sure it's a satisfactorily strong gradient.
                            if (grad[sY, sX] > minGrad):
                                gradWeight.append(grad[sY, sX])
                                gradPoint.append([sX, sY])

                # Relax our threshold if we don't have enough points.
                minGrad = minGrad * 0.5

            a = -999
            b = -999

            # Now perform the linear regression (if we can).
            if (len(gradPoint) > 1):
                a, b = weightedLinearRegression(gradPoint, gradWeight, 0)

            m = a

            # Okay, we have a boundary.
            #
            # NOTE: There are actually some situations where 'a' can be NaN (not a number).
            # Make sure that's not the case.
            if (a != -999 and not(math.isnan(a)) and not(math.isnan(math.atan(a))) and a != None):
                try:
                    a = math.atan(a)

                    frontAngle[y:y+size, x:x+size] = a * (180 / math.pi)
                except:
                    frontAngle[y:y+size, x:x+size] = -999
            else:
                # No front here? Enter a numerical value that represents N/A

                frontAngle[y:y+size, x:x+size] = -999

    return frontAngle


# This is a function that calculates the layer mean value of a metric given a
# series of altitudes (sorted from lowest to highest). This is an improvement 
# over the old method for calculating mixed layer parcel properties.

def continuumLayerAverage(x, z, bottom, top):
    xTop = interpolate(x, z, top, reverse=False)
    xBottom = interpolate(x, z, bottom, reverse=False)

    deltaTop = None
    deltaBottom = None

    subAverageTop = 0
    subAverageBottom = 0

    subAverage = []
    delta = []

    if (bottom == top):
        return ((xTop + xBottom) / 2)

    for i in range(0, len(z)):
        if (z[i] >= bottom and deltaBottom == None):
            if (z[i] == bottom):
                deltaBottom = 0
            else:
                deltaBottom = z[i] - bottom
                m = (xBottom - x[i]) / (bottom - z[i])
                b = xBottom - m * bottom
                subAverageBottom = (m / 2) * (bottom + z[i]) + b
        elif (z[i] >= top and deltaTop == None):
            if (z[i] == top):
                deltaTop = z[i] - z[i-1]
                m = (x[i] - x[i-1]) / (z[i] - z[i-1])
                b = x[i] - m * z[i]
                subAverageTop = (m / 2) * (z[i] + z[i-1]) + b
            else:
                deltaTop = top - z[i-1]
                m = (xTop - x[i-1]) / (top - z[i-1])
                b = x[i-1] - m * z[i-1]
                subAverageTop = (m / 2) * (top + z[i-1]) + b
        elif (z[i] >= bottom and z[i] <= top):
            delta.append(z[i] - z[i-1])
            m = (x[i] - x[i-1]) / (z[i] - z[i-1])
            b = x[i] - m * z[i]
            a = (m / 2) * (z[i] + z[i-1]) + b
            subAverage.append(a)

    theResult = 0

    if (len(subAverage) == 0):
        theResult = (xTop + xBottom) / 2
    else:
        theSum = 0
        weight = 0

        theSum = theSum + subAverageBottom * deltaBottom
        weight = weight + deltaBottom

        for i in range(0, len(delta)):
            theSum = theSum + subAverage[i] * delta[i]
            weight = weight + delta[i]

        theSum = theSum + subAverageTop * deltaTop
        weight = weight + deltaTop

        theResult = theSum / weight

    return theResult


# This function takes in the geopotential height grid and then uses that information
# to determine the pressure level at ground level (not sea level). This works because
# negative height values are logged for pressures that occur below ground level.

def getSurfacePressure(height):
    surfacePressure = np.zeros([len(height.data[0][:,0]), len(height.data[0][0,:])])

    converted = np.array(height.data)

    for i in range(0, len(surfacePressure[:, 0])):
        for j in range(0, len(surfacePressure[0, :])):
            z = interpolate(height.level, converted[:, i, j], 0)

            surfacePressure[i, j] = z

    return surfacePressure


# Algorithm to estimate initiation probability P0 given the surface pressure, a sounding, 
# vertical velocity, and uncertainty terms (for temperature and vertical velocity). FUN 
# FACT: This algorithm took over a month to get in working order (several intermittent 
# and devilishly subtle bugs needed to be eliminated).
#
# In version 1.0.2, this algorithm now accounts for the vertical motions present on
# model forecast soundings and incorporates an improved method for estimating the
# properties of a 100 mb mixed layer air parcel.

def getOverallProbability(surfacePressure, rawData, w, sigmaT, sigmaW):
    rawArray = np.array(rawData)

    # These variables will store the critical values of virtual temperature and vertical
    # velocity.
    T_hat = -999
    w_hat = -999

    w_obs = w

    pressure = rawArray[:, 0]
    height = rawArray[:, 1]
    temperature = rawArray[:, 2]
    dewpoint = rawArray[:, 3]

    bottomP = surfacePressure
    topP = surfacePressure - 100
    
    bottomZ = 0
    topZ = interpolate(height, pressure, topP, reverse=True)

    PML = continuumLayerAverage(pressure, height, bottomZ, topZ)
    zML = (bottomZ + topZ) / 2
    DML = continuumLayerAverage(dewpoint, height, bottomZ, topZ)
    TML = continuumLayerAverage(temperature, height, bottomZ, topZ)

    # Calculate the virtual temperature of the mixed-layer parcel...     
    T = virtualTemperature(TML, DML, "K", "K")

    # ...and remember that value.
    T_initial = T
    T_parcel  = T
    T_trace   = []
    P_trace   = []
        
    # Remember the vertical velocity we were given.
    w_initial = w
    
    # We need to know what the parcel's mixing ratio is to know if it is saturated or not.
    r = mixingRatio(DML, PML*100)
    
    # Initial value for a height increment (will be overwritten later).
    deltaZ = 0.1

    # Increase/decrease temperature by this much until we find the temperature that just
    # barely causes the mixed layer parcel to reach the EL.
    deltaT = 0.1

    # Increase/decrease vertical velocity by this much until we find the temperature that 
    # just barely causes the mixed layer parcel to reach the EL.
    deltaW = 0.01
    
    # Time increment (in seconds), ideally should be infinitesimally small, but even this
    # large number is computationally expensive to use.
    timeStep = 1.0
    
    # Air parcel starts off being unsaturated.
    isSaturated = False
    
    # Algorithm to estimate where the tropopause is. The process involves looking for a
    # contiguous region where the temperature is steadily increasing with height, which is
    # an indication we're in the stratosphere.
    #
    # NOTE: The astute reader might ask "Why not just use EL or LFC?". The problem with
    # using standard EL is this height will change as we manipulate the virtual
    # temperature of the mixed layer air parcel. By not fixing this height, we can
    # actually run into problems, especially in environments characterized by shallow
    # lapse rates. Using LFC runs into the same problem, but it also turns out that
    # reaching the LFC doesn't necessarily mean the parcel will rise to the tropopause.
    # Just because an air parcel becomes positively buoyant at some point doesn't mean it 
    # will stay positively buoyant throughout the troposphere.
    #
    # NOTE: We cannot just simply look for a single temperature reading that increases
    # with height and call that the tropopause. If we did this, inversions (either from
    # an elevated mixed layer or mid-level subsidence) would be misidentified as the
    # tropopause.
    #
    # NOTE: If there is a tropopause fold, there will be a layer of tropospheric air on
    # top of stratospheric air. This will show up as a large layer of increasing air
    # temperature, followed by a large layer of decreasing air temperature, followed by
    # another layer of increasing air temperature. This could give us two tropopauses, but
    # we should realistically take the lowest one since an air parcel is unlikely to rise
    # above such a stable layer.
    #
    # NOTE: There is also a minimum and maximum value for estimated tropopause height.
    # This is a safeguard against two contingencies. First, a massive inversion near
    # ground level can be misidentified as the tropopause if it is deep enough. In this
    # case, the tropopause height is set to the level of non-divergence (around 500 mb or
    # a height of 6 km above sea level). On rare occasions, the tropopause can be lower
    # than this, but that would normally be found in a polar air mass. However, there
    # have been documented cases of severe weather occurring with anomalously low
    # tropopause heights. Such events are exceedingly rare, meaning this algorithm will
    # work for about 99.9% of all situations. Second, if we have a maritime air mass with
    # a high tropopause and shallow environmental lapse rates, the tropopause height will
    # be capped at 10 km. Otherwise, an unrealistically high virtual temperature and
    # vertical velocity will be required to reach the tropopause. If an air parcel reaches
    # a height of 10 km, chances are a convective cell has formed, even if there is an
    # extra 5 km of troposphere above it.
    EL = 0
    EL_strikes = 0
    EL_max_strikes = 3
    EL_flag = False
    
    lastT = -999
    
    n = 0
        
    EL_z = -1
    
    # Step up through the troposphere.
    for z in range(0, len(rawData)):

        # Check if this temperature is higher than the last temperature we queried. If it
        # is, the temperature is increasing with height.
        if (rawData[z][2] > lastT and lastT > -999):

            # Not all temperature increases should be treated the same. 
            EL_strikes = EL_strikes + ((rawData[z][2]) / (lastT))

            # Remember this height incase this really is the tropopause.
            if (EL_z < 0):
                EL_z = z

            # We found a contiguous layer of increasing temperatures.
            if (EL_strikes >= EL_max_strikes):
                EL = rawData[EL_z][1]
                EL_flag = True
                break
        else:
            EL_z = -1         # Reset this variable if the air is cooling with height.
            EL_strikes = 0
        
        # Remember the last temperature we looked at.
        lastT = rawData[z][2]
        
    # If we could not satisfy the criteria of a contiguous layer of increasing
    # temperatures, take the last layer of increasing temperature we found and call that
    # the tropopause height.
    if (not(EL_flag)):
        EL = rawData[EL_z][1]
        
    # Make sure we have a reasonable height for air parcels to reach. If the tropopause
    # height is too high, we're probably in a maritime environment with shallow lapse
    # rates. If it's too low, we probably misidentified a strong low-level inversion as 
    # the tropopause.
    EL = min(EL, 10000)
    EL = max(EL, 6000)
        
    # Set the initial height and pressure to that of the mixed layer air parcel.
    z = zML
    P = PML
    
    # This variable will prevent oscillating behavior in the below loop.
    zeroFlag = False
    
    # Set the initial vertical height to what the model predicted.
    w = w_obs
        
    # This variable will tell us if the ambient environment is already unstable for a
    # mixed layer parcel traveling at the given vertical velocity.
    autoConvective = True

    # IMPORTANT NOTE: This algorithm is NOT the same thing as a simple parcel trace you'd
    # see on a sounding. On a simple parcel trace, you always assume the parcel is moving
    # upward. We make no such assumption here.

    # This first loop will find the critical value of virtual temperature.
    while T_hat <= -999:
        # A linear interpolation algorithm to infer what the temperature, dewpoint,
        # pressure, and background vertical velocity are at any given height.
        bestUpper = 0
        bestLower = 0
        
        for h in range(1, len(rawData)):
            if (rawData[h][1] > z and rawData[h][2] > -999 and rawData[h-1][2] > -999):
                bestUpper = h
                break
            
        for h in range(bestUpper - 1, 0, -1):
            if (rawData[h][2] > -999):
                bestLower = h
                break
            
        dT_dz = (rawData[bestUpper][2] - rawData[bestLower][2]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dD_dz = (rawData[bestUpper][3] - rawData[bestLower][3]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dP_dz = (rawData[bestUpper][0] - rawData[bestLower][0]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dW_dz = (rawData[bestUpper][6] - rawData[bestLower][6]) / (rawData[bestUpper][1] - rawData[bestLower][1])

        T = dT_dz * (z - rawData[bestLower][1]) + rawData[bestLower][2]
        D = dD_dz * (z - rawData[bestLower][1]) + rawData[bestLower][3]
        P = dP_dz * (z - rawData[bestLower][1]) + rawData[bestLower][0]
        omega = dW_dz * (z - rawData[bestLower][1]) + rawData[bestLower][6]

        # Combine temperature and dewpoint into a single metric.
        T = virtualTemperature(T, D, "K", "K")
        
        # For debugging purposes
        T_trace.append(T_parcel)
        P_trace.append(P)
        
        # If the air parcel is saturated, decrease the parcel's temperature by the moist
        # adiabatic (pseudoadiabatic) lapse rate. Otherwise, we'll use the dry adiabatic
        # lapse rate.
        if (isSaturated):
            T_parcel = T_parcel - (moistLapseRate(T_parcel, P, "K", "hPa") / 1000) * deltaZ
        else:
            T_parcel = T_parcel - 0.0098 * deltaZ

            # Temperature has changed, so we have a new value for saturation mixing ratio.
            saturationMixingRatio = mixingRatio(T_parcel, P*100)
            
            # Is the parcel now saturated?
            if (r >= saturationMixingRatio):
                isSaturated = True
                
                
        # Calculate buoyant force and the consequent change in vertical velocity
        w = w + 9.8 * ( (T_parcel / T) - 1 ) * timeStep + omega

        # Calculate the change in height from our new vertical velocity. NOTE: We need to
        # remember what the height increment is for the above lapse rate calculations.
        deltaZ = w * timeStep
        z = z + deltaZ
                
        # The parcel has reached the tropopause.
        if (z > EL):

            # Was the atmosphere already unstable? If so, cool the mixed layer air parcel
            # until it becomes stable.
            if (autoConvective):
                n = n - 1

                T = virtualTemperature(TML, DML, "K", "K")
                T_initial = T + n * deltaT
                T_parcel  = T + n * deltaT
                T_trace   = []
                P_trace   = []

                # MERCY RULE: If the critical temperature is more than three standard
                # deviations away from "observation", there's little point in continuing.
                if (abs(T_initial - T) / sigmaT >= 3):
                    T_hat = T_initial
                    break

                # Reset everything for the next cycle in the loop.
                w = w_obs
                
                isSaturated = False
                
                z = zML
                P = PML

                zeroFlag = True

                continue

            # If the atmosphere wasn't initially unstable, but now is for this value of
            # virtual temperature, we've found the critical value for virtual temperature.
            else:
                T_hat = T_initial
                break

        # The parcel is now stable and won't go any higher. So, we need a higher virtual
        # temperature to reach the tropopause.
        elif (z < 0 or w < 0):            
            n = n + 1
            
            T = virtualTemperature(TML, DML, "K", "K")
            T_initial = T + n * deltaT
            T_parcel  = T + n * deltaT
            T_trace   = []
            P_trace   = []

            # MERCY RULE: If the critical temperature is more than three standard
            # deviations away from "observation", there's little point in continuing.
            if (abs(T_initial - T) / sigmaT >= 3):
                T_hat = T_initial
                break

            # Reset everything for the next cycle in the loop.
            w = w_obs

            isSaturated = False
            
            # THIS IS AN IMPORTANT LINE OF CODE. Otherwise, we could accidentally create
            # an infinite loop.
            autoConvective = False
            
            z = zML
            P = PML
            
            continue
           

    # Reset everything for the next loop, which will determine the critical value for 
    # vertical velocity.

    z = zML
    P = PML
    T = virtualTemperature(TML, DML, "K", "K")
    T_initial = T
    T_parcel  = T

    w = w_obs

    isSaturated = False
    
    n = 0
    
    autoConvective = True
    
    zeroFlag = False
    
    while w_hat <= -999:
        # Same linear interpolation algorithm as in the first loop
        bestUpper = 0
        bestLower = 0
        
        for h in range(1, len(rawData)):
            if (rawData[h][1] > z and rawData[h][2] > -999 and rawData[h-1][2] > -999):
                bestUpper = h
                break
            
        for h in range(bestUpper - 1, 0, -1):
            if (rawData[h][2] > -999):
                bestLower = h
                break
        
        dT_dz = (rawData[bestUpper][2] - rawData[bestLower][2]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dD_dz = (rawData[bestUpper][3] - rawData[bestLower][3]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dP_dz = (rawData[bestUpper][0] - rawData[bestLower][0]) / (rawData[bestUpper][1] - rawData[bestLower][1])
        dW_dz = (rawData[bestUpper][6] - rawData[bestLower][6]) / (rawData[bestUpper][1] - rawData[bestLower][1])

        T = dT_dz * (z - rawData[bestLower][1]) + rawData[bestLower][2]
        D = dD_dz * (z - rawData[bestLower][1]) + rawData[bestLower][3]
        P = dP_dz * (z - rawData[bestLower][1]) + rawData[bestLower][0]
        omega = dW_dz * (z - rawData[bestLower][1]) + rawData[bestLower][6]

        T = virtualTemperature(T, D, "K", "K")
        
        # This is also the same.
        if (isSaturated):
            T_parcel = T_parcel - (moistLapseRate(T_parcel, P, "K", "hPa") / 1000) * deltaZ
        else:
            T_parcel = T_parcel - 0.0098 * deltaZ
            saturationMixingRatio = mixingRatio(T_parcel, P*100)
            
            if (r >= saturationMixingRatio):
                isSaturated = True
                
        w = w + 9.8 * ( (T_parcel / T) - 1 ) * timeStep + omega
        deltaZ = w * timeStep
        z = z + deltaZ

        # Same idea as before, but now apply the changes to vertical velocity instead of
        # temperature. If the atmosphere was already unstable, bring the vertical velocity
        # down until it becomes stable.
        if (z > EL):
            if (autoConvective):
                n = n - 1
                
                w = w_obs + n * deltaW
                w_initial = w
                
                # MERCY RULE: If the critical velocity is more than three standard
                # deviations away from "observation", there's little point in continuing.
                if (abs(w - w_obs) / sigmaW >= 3):
                    w_hat = w_initial
                    break

                # Reset everything for the next cycle in the loop.
                T = virtualTemperature(TML, DML, "K", "K")
                T_initial = T
                T_parcel  = T
                
                isSaturated = False
                
                z = zML
                P = PML

                zeroFlag = True
                
                continue
            else:
                w_hat = w_initial
                break

        # The parcel is now stable and won't go any higher. So, we need a higher velocity
        # to reach the tropopause.
        elif (z < 0 or w < 0):
            if (n < 0 or (n >= 0 and zeroFlag)):
                w_hat = w_initial
                break
            
            n = n + 1
            
            w = w_obs + n * deltaW
            w_initial = w

            # MERCY RULE: If the critical velocity is more than three standard
            # deviations away from "observation", there's little point in continuing.
            if (abs(w - w_obs) / sigmaW >= 3):
                w_hat = w_initial
                break
            
            # Reset everything for the next cycle in the loop.
            T = virtualTemperature(TML, DML, "K", "K")
            T_initial = T
            T_parcel  = T
            
            isSaturated = False
            
            autoConvective = False
            
            z = zML
            P = PML
            
            continue
                
    w = w_obs
    T = virtualTemperature(TML, DML, "K", "K")
            

    # Calculate the probability of the actual virtual temperature being high enough to
    # initiate convection.
    temperatureProbability = gaussianProbability(T_hat, sigmaT, -1000.0, T, 0.1)

    # Calculate the probability of the actual vertical velocity being high enough to
    # initiate convection.
    forcingProbability     = gaussianProbability(w_hat, sigmaW, -1000.0, w, 0.01)

    # Now calculate the mean between those two values. NOTE: Originally wanted a mean
    # weighted to the normalized uncertainties, but this presented too many problems in
    # practice. As an example, what do you do with a negative vertical velocity in a
    # coefficient that's supposed to be positive? There may be a way to accomplish this
    # using some sort of scale analysis, but vertical velocity (even in a global model)
    # can cover multiple orders of magnitude.
    overallProbability = 0.5 * (temperatureProbability + forcingProbability)

    return overallProbability


# Trigonometry functions in terms of degrees instead of radians.

def sin(x):
    return (math.sin(x * (math.pi / 180)))


def cos(x):
    return (math.cos(x * (math.pi / 180)))


def tan(x):
    return (math.tan(x * (math.pi / 180)))


# This is the statistically derived graphical discriminate function that
# estimates storm mode using the spacing parameter (see original publication
# for additional details).

def f(D, delta_s_bar):
    rI = 17.338270116131493
    rL = -3.0404932831625557

    sigmaI = 33.39769464119628
    sigmaL = 15.152093030165432

    L = 1 - gaussianProbability(rL, sigmaL, -1000, D, 0.1, coefficient=2)
    I =     gaussianProbability(rI, sigmaI, -1000, D, 0.1, coefficient=2)

    return (L / (I + L))


# Method for loading a .grb file from its raw binaries. Note that the
# current version of the pygrib module can only load data from files that
# were already downloaded to your computer. This function has no such
# requirements; it only needs the raw data of the file's contents.

def loadGRIB(x):
    messageList = []

    buffer = b''
    i = 0

    segment = x.split(b'7777GRIB')

    for i in range(0, len(segment)):
        if (len(segment[i]) > 0):
            buffer = segment[i]

            if (buffer[0:4] != b'GRIB'):
                buffer = b'GRIB' + buffer

            if (buffer[-4:] != b'7777'):
                buffer = buffer + b'7777'

            test = pygrib.fromstring(buffer)
            messageList.append(test)

    return messageList


# Function to make loading model data easier. The URLs are slightly different for each
# individual model, which made having a simplification highly desirable.

def loadModel(theModel, year, month, day, runHour, forecastHour, memberID=0):
    sourceURL = ""
    
    # The 'memberID' argument is ignored if we're not looking at an ensemble. If
    # memberID is 0, then we're actually querying the ensemble mean. Otherwise, we're
    # querying a specific member's output.

    if (theModel == "GEFS"):
        if (memberID < 0 or memberID > 30):
            raise ValueError("Invalid member ID: " + str(memberID))
            
        if (memberID == 0):
            sourceURL = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/gens/prod/gefs."+str(year)+fitLength(month, 2)+fitLength(day, 2)+"/"+fitLength(runHour, 2)+"/atmos/pgrb2ap5/geavg.t"+fitLength(runHour, 2)+"z.pgrb2a.0p50.f" + fitLength(forecastHour, 3)
        else:
            sourceURL = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/gens/prod/gefs."+str(year)+fitLength(month, 2)+fitLength(day, 2)+"/"+fitLength(runHour, 2)+"/atmos/pgrb2ap5/gep"+fitLength(memberID, 2)+".t"+fitLength(runHour, 2)+"z.pgrb2a.0p50.f" + fitLength(forecastHour, 3)
        
    if (theModel == "GFS"):
        sourceURL = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/gfs."+str(year)+fitLength(month, 2)+fitLength(day, 2)+"/"+fitLength(runHour, 2)+"/atmos/gfs.t"+fitLength(runHour, 2)+"z.pgrb2.0p50.f" + fitLength(forecastHour, 3)
        
    if (theModel == "NAM"):
        sourceURL = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/nam/prod/nam."+str(year)+fitLength(month, 2)+fitLength(day, 2)+"/nam.t"+fitLength(runHour, 2)+"z.awphys"+fitLength(forecastHour, 2)+".tm00.grib2"
        
        
    outFile = theModel + "-" + str(year) + fitLength(month, 2) + fitLength(day, 2) + "-FH" + fitLength(forecastHour, 3) + ".grb2"

    if (theModel == "GEFS"):
        outFile = theModel + "-" + str(memberID) + "-" + str(year) + fitLength(month, 2) + fitLength(day, 2) + "-FH" + fitLength(forecastHour, 3) + ".grb2"
    
    if (os.path.exists(outFile)):
        pointer = open(outFile, 'rb')
        rawData = pointer.read()
    elif (sourceURL != "" and not(os.path.exists(outFile))):
        rawData = loadResource(sourceURL)
    
        # If get a network error, the length of rawData will be less than 1000 characters.
        if (len(rawData) <= 1000):
            raise ValueError("Encountered an unexpected response: ", str(rawData))
    
    elif (sourceURL == ""):
        raise ValueError("Model data for " + outFile[:-5] + " was not found")
    

    # Use the pygrib module to open the file we just downloaded.
    modelData = loadGRIB(rawData)
    
    return modelData


global settings

settings = {}


# Method for cleaning up any unnecessary spaces or insignificant characters.

def filter(x):
    start = -1
    end = -1

    for i in range(0, len(x)):
        if (ord(x[i]) > 0x20):
            start = i
            break

    for i in range(len(x)-1, -1, -1):
        if (ord(x[i]) > 0x20):
            end = i+1
            break

    return x[start:end]


# Method for loading and validating the configuration file "config.txt"

def loadConfiguration():
    print("Loading configuration file...")
    pointer = open("config.txt", 'r')
    raw = pointer.read()
    line = raw.split('\n')

    parameterList = ["run-date", "run-hour", "forecast-hour", "latitude-domain", "longitude-domain", "save-grid", "save-map", "resolution", "parent"]
    fileList = []
    errorList = []

    for i in range(0, len(line)):
        linePrefix = f"Line {i+1}: "

        if (len(line[i]) == 0):
            continue

        buffer = line[i].split(':')

        if (len(buffer) != 2):
            print(linePrefix + "Missing or extra colon separating parameter and value")

        parameterName = filter(buffer[0])
        parameterValue = filter(buffer[-1])

        fileList.append(parameterName)

        if (not(parameterName in parameterList)):
            errorList.append(linePrefix + 'Unrecognized parameter "'+parameterName+'"')

        dateKey = ["month", "day", "year"]
        rawDate = []

        if ("run-date" in parameterName):
            settings['run-date'] = parameterValue

            if ("/" in parameterValue):
                rawDate = parameterValue.split('/')
            elif ("-" in parameterValue):
                rawDate = parameterValue.split('-')
            elif ("." in parameterValue):
                rawDate = parameterValue.split('.')
            elif (" " in parameterValue):
                rawDate = parameterValue.split(' ')
            else:
                if (len(parameterValue) != 8):
                    errorList.append(linePrefix + 'Invalid time stamp format "'+parameterValue+'"')

                rawDate.append(parameterValue[0:2])
                rawDate.append(parameterValue[2:4])
                rawDate.append(parameterValue[4:8])

            if (len(rawDate) != 3):
                 errorList.append(linePrefix + 'Invalid date "'+parameterValue+'"')

            badDate = False

            for j in range(0, len(rawDate)):
                try:
                    settings[dateKey[j]] = int(rawDate[j])
                except:
                    errorList.append(linePrefix + 'Month, day, and year values must be integers')
                    badDate = True

            if (badDate):
                continue

            if (settings['month'] < 1 or settings['month'] > 12):
                errorList.append(linePrefix + 'Invalid month (must be between 1 and 12)')
            else:
                month_day = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

                if (settings['year'] % 4 == 0):
                    month_day[2] = 29

                if (settings['day'] < 1 or settings['day'] > month_day[settings['month']]):
                    errorList.append(linePrefix + 'Invalid day (must be between 1 and '+str(month_day[settings['month']])+') for the given month')
        elif ("run-hour" in parameterName):
            try:
                settings['run-hour'] = int(parameterValue)

                if (settings['run-hour'] % 6 != 0 or settings['run-hour'] >= 24 or settings['run-hour'] < 0):
                    errorList.append(linePrefix + 'Run hour value must be 0, 6, 12, or 18')
            except:
                errorList.append(linePrefix + 'Run hour value must be an integer')
                continue
        elif ("forecast-hour" in parameterName):
            if ((parameterValue[0] != "(" and parameterValue[0] != "[") or (parameterValue[-1] != ")" and parameterValue[-1] != "]")):
                errorList.append(linePrefix + 'Missing parentheses or brackets on bounds')

            hourBuffer = parameterValue[1:-1].split(',')

            try:
                if (len(hourBuffer) == 1):
                    settings['forecast-hour'] = [int(filter(hourBuffer[0])), int(filter(hourBuffer[0]))+2]

                    if (parameterValue[0] == "(" or parameterValue[-1] == ")"):
                        errorList.append(linePrefix + 'Must use square brackets if accessing a single forecast hour')
                elif (len(hourBuffer) == 2):
                    settings['forecast-hour'] = [int(filter(hourBuffer[0])), int(filter(hourBuffer[1]))]
                elif (len(hourBuffer) >= 3):
                    errorList.append(linePrefix + 'Too many values in forecast hour entry')
                    continue
            except:
                errorList.append(linePrefix + 'Run hour bounds must be integers')
                continue

            if (settings['forecast-hour'][0] > settings['forecast-hour'][1]):
                errorList.append(linePrefix + 'Run hour bounds are out of order (smallest number must appear before the biggest number)')

            if (settings['forecast-hour'][0] > 240 or settings['forecast-hour'][1] > 240):
                errorList.append(linePrefix + 'Invalid forecast hour value (must be between 0 and 240)')

            if (settings['forecast-hour'][0] < 0 or settings['forecast-hour'][1] < 0):
                errorList.append(linePrefix + 'Invalid forecast hour value (must be between 0 and 240)')

            if (settings['forecast-hour'][0] % 3 != 0):
                errorList.append(linePrefix + 'Invalid forecast hour value (first value must be a multiple of 3)')

            if (parameterValue[0] == "("):
                settings['forecast-hour'][0] = settings['forecast-hour'][0] + 3

            if (parameterValue[-1] == "]"):
                settings['forecast-hour'][1] = settings['forecast-hour'][1] + 1
        elif ("latitude-domain" in parameterName):
            if ((parameterValue[0] != "(" and parameterValue[0] != "[") or (parameterValue[-1] != ")" and parameterValue[-1] != "]")):
                errorList.append(linePrefix + 'Missing parentheses or brackets on bounds')

            boundBuffer = parameterValue[1:-1].split(',')

            try:
                if (len(boundBuffer) == 2):
                    settings['latitude-domain'] = [float(filter(boundBuffer[0])), float(filter(boundBuffer[1]))]
                else:
                    errorList.append(linePrefix + 'Two values needed in latitude domain entry')
                    continue
            except:
                errorList.append(linePrefix + 'Latitude domain bounds must be numbers')
                continue

            if (parameterValue[0] == "("):
                settings['latitude-domain'][0] = settings['latitude-domain'][0] + 0.5
            if (parameterValue[-1] == "]"):
                settings['latitude-domain'][1] = settings['latitude-domain'][1] + 0.5

            if (settings['latitude-domain'][0] % 1 != 0 and settings['latitude-domain'][0] % 1 != 0.5):
                errorList.append(linePrefix + 'Latitude domain bounds must be multiples of 0.5')
            if (settings['latitude-domain'][1] % 1 != 0 and settings['latitude-domain'][1] % 1 != 0.5):
                errorList.append(linePrefix + 'Latitude domain bounds must be multiples of 0.5')

            if (settings['latitude-domain'][0] < -90 or settings['latitude-domain'][0] > 90 or settings['latitude-domain'][1] < -90 or settings['latitude-domain'][1] > 90):
                errorList.append(linePrefix + 'Invalid latitude values (must be between -90 and 90)')

            if (abs(settings['latitude-domain'][1] - settings['latitude-domain'][0]) < 8):
                errorList.append(linePrefix + 'Latitude domain is too small (must span at least 8 degrees)')

            if (settings['latitude-domain'][1] < settings['latitude-domain'][0]):
                errorList.append(linePrefix + 'Latitude bounds are out of order (smallest value must appear first)')
        elif ("longitude-domain" in parameterName):
            if ((parameterValue[0] != "(" and parameterValue[0] != "[") or (parameterValue[-1] != ")" and parameterValue[-1] != "]")):
                errorList.append(linePrefix + 'Missing parentheses or brackets on bounds')

            boundBuffer = parameterValue[1:-1].split(',')

            try:
                if (len(boundBuffer) == 2):
                    settings['longitude-domain'] = [float(filter(boundBuffer[0])), float(filter(boundBuffer[1]))]

                    if (settings['longitude-domain'][0] < 0):
                        settings['longitude-domain'][0] = 360 + settings['longitude-domain'][0]
                    if (settings['longitude-domain'][1] < 0):
                        settings['longitude-domain'][1] = 360 + settings['longitude-domain'][1]
                else:
                    errorList.append(linePrefix + 'Two values needed in longitude domain entry')
                    continue
            except:
                errorList.append(linePrefix + 'Longitude domain bounds must be numbers')
                continue

            if (parameterValue[0] == "("):
                settings['longitude-domain'][0] = settings['longitude-domain'][0] + 0.5
            if (parameterValue[-1] == "]"):
                settings['longitude-domain'][1] = settings['longitude-domain'][1] + 0.5

            if (settings['longitude-domain'][0] % 1 != 0 and settings['longitude-domain'][0] % 1 != 0.5):
                errorList.append(linePrefix + 'Longitude domain bounds must be multiples of 0.5')
            if (settings['longitude-domain'][1] % 1 != 0 and settings['longitude-domain'][1] % 1 != 0.5):
                errorList.append(linePrefix + 'Longitude domain bounds must be multiples of 0.5')

            if (settings['longitude-domain'][0] < 0 or settings['longitude-domain'][0] > 360 or settings['longitude-domain'][1] < 0 or settings['longitude-domain'][1] > 360):
                errorList.append(linePrefix + 'Invalid longitude values (must be between -180 and 180 or between 0 and 360)')

            if (abs(settings['longitude-domain'][1] - settings['longitude-domain'][0]) < 8):
                errorList.append(linePrefix + 'Longitude domain is too small (must span at least 8 degrees)')

            if (settings['longitude-domain'][1] < settings['longitude-domain'][0]):
                errorList.append(linePrefix + 'Longitude bounds are out of order (smallest value must appear first)')
        elif ("save" in parameterName):
            if (parameterValue.upper() == "TRUE" or parameterValue.upper() == "YES" or parameterValue.upper() == "Y"):
                settings[parameterName] = True
            elif (parameterValue.upper() == "FALSE" or parameterValue.upper() == "NO" or parameterValue.upper() == "N"):
                settings[parameterName] = False
            else:
                errorList.append(linePrefix + 'Invalid boolean value (must be "true" or "false")')
        elif ("resolution" in parameterName):
            try:
                settings['resolution'] = int(parameterValue)

                if (settings['resolution'] % 2 != 0):
                    errorList.append(linePrefix + 'Resolution parameter must be a multiple of 2 (a value of 2 or 4 works best)')

                if (settings['resolution'] > 16):
                    errorList.append(linePrefix + 'Resolution parameter cannot be larger than 16 (a value of 2 or 4 works best)')

                if (settings['resolution'] <= 0):
                    errorList.append(linePrefix + 'Resolution parameter must be a positive number')
            except:
                errorList.append(linePrefix + 'Resolution value must be an integer')
                continue
        elif ("parent" in parameterName):
            if (parameterValue[0] != '"' or parameterValue[-1] != '"'):
                errorList.append(linePrefix + 'Parent value must be a string encased in quotes ""')
                continue
            else:
                settings['parent'] = parameterValue[1:-1]

    for i in range(0, len(parameterList)):
        if (not(parameterList[i] in settings) and not(parameterList[i] in fileList)):
            errorList.append(linePrefix + 'Required parameter "'+parameterList[i]+'" is missing')

    if (len(errorList) > 0):
        numErrors = len(errorList)

        if (numErrors == 1):
            print(f'{numErrors} error occurred when attempting to read "config.txt":')
        else:
            print(f'{numErrors} errors occurred when attempting to read "config.txt":')

        for i in range(0, len(errorList)):
            print("    "+errorList[i])

        sys.exit()

    if (not(settings['save-grid']) and not(settings['save-map'])):
        print("\aWARNING: Forecast output will not be saved in any format.")

    if (settings['resolution'] != 2 and settings['resolution'] != 4):
        print("\aWARNING: Unrecommended resolution value used (a value of 2 or 4 works best)")


loadConfiguration()

global data

data = {}

year  = settings['year']
month = settings['month']
day   = settings['day']
hour  = settings['run-hour']

startHour = settings['forecast-hour'][0]
endHour   = settings['forecast-hour'][1]

# Lower values of resolution will look at smaller subsections. Higher values will look at
# larger subsections. Larger subsections run faster, but may not be as useful to work
# with.
resolution = settings['resolution']

# This is an empirically estimated quantity (see original publication for details).
gain = 1.414

print()
print("Running model for:")
print(f"GFS/GEFS initialized on {month}/{day}/{year} at {fitLength(hour, 2)}Z")

numTimes = 0

if (endHour % 3 == 1):
    print(f"FH{fitLength(startHour, 3)} - FH{fitLength(endHour - 1, 3)}")
    print(f"(including FH{fitLength(endHour - 1, 3)})")
    numTimes = (endHour - startHour) // 3 + 1
else:
    print(f"FH{fitLength(startHour, 3)} - FH{fitLength(endHour - 3, 3)}")
    numTimes = (endHour - startHour) // 3

print(f"Latitudes:  {formatLatitude(settings['latitude-domain'][0])} to {formatLatitude(settings['latitude-domain'][1] - 0.5)}")
print(f"Longitudes: {formatLongitude(settings['longitude-domain'][0])} to {formatLongitude(settings['longitude-domain'][1] - 0.5)}")


# Indices that point to specific domains of latitude and longitude we want.
lowerLatIndex = int(-2 * (settings['latitude-domain'][1] - 90))
upperLatIndex = int(-2 * (settings['latitude-domain'][0] - 90))

lowerLonIndex = int(2 * (settings['longitude-domain'][0]))
upperLonIndex = int(2 * (settings['longitude-domain'][1]))

subLatIndex = [lowerLatIndex, upperLatIndex]
subLonIndex = [lowerLonIndex, upperLonIndex]

print(f"Resultant grid (time x latitude x longitude): {numTimes} x {abs(upperLatIndex-lowerLatIndex)} x {abs(upperLonIndex-lowerLonIndex)}")


day = settings['day']
month = settings['month']
year = settings['year']
originalDay = day
originalMonth = month
originalYear = year

stampPrefix = fitLength(settings['month'], 2) + "-" + fitLength(settings['day'], 2) + "-" + fitLength(settings['year'], 4) + " " + fitLength(settings['run-hour'], 2) + "Z Run "
parentDirectory = settings['parent']

if (parentDirectory[-1] != "/"):
    parentDirectory = parentDirectory + '/'

# Empirically calculated arrays for storms that don't initiate along boundaries
stochasticProbability = np.array([0.0, 0.017519378876249807, 0.035038757752499615, 0.05255813662874942, 0.07007751550499923, 0.08759689438124903, 0.10511627325749884, 0.12263565213374863, 0.14015503100999846, 0.15767440988624823, 0.17519378876249805, 0.19271316763874785, 0.21023254651499768, 0.22775192539124747, 0.24527130426749727, 1.0])
stochastic = np.array([0.0, 0.0, 0.2617705523642253, 0.6249348127478669, 0.6761838490944435, 0.7368873524561085, 0.7772581954795299, 0.8131449754358315, 0.8425603599991879, 0.8552418196074766, 0.8701798526920211, 0.8798144722221290, 0.8900147408772756, 0.8935246802040249, 0.9006994797412474, 1.0])

if (settings['save-map']):
    import matplotlib.pyplot as plt
    import cartopy.crs as ccrs
    import cartopy.feature as cf
    from matplotlib.colors import *


print()

# Empty lists that will be filled by the below loop
groundPressure = []
temperatureBuffer = []
humidityBuffer = []
omegaBuffer = []
heightBuffer = []
uBuffer = []
vBuffer = []

pressureField = None

# Regarding the below loop:
# If member is -1, we retrieve the GFS instead GEFS
# If member is 0, we retrieve the GEFS ensemble mean
# If member is 1 or higher, we retrieve that specific member of GEFS

FH = startHour
originalHour = startHour

while (FH < endHour):
    member = -1

    while (member <= 30):
        # A nice little progress display
        print("\rLoading member #" + str(member+1) + ", FH" + fitLength(FH, 3) + "               ", end="", flush=True)

        if (member == -1):
            modelData = None

            try:
                modelData = loadModel("GFS", year, month, day, hour, FH)
            except ValueError:
                print(f"\n\aERROR: Model data for the {fitLength(hour, 2)}Z run on {month}/{day}/{year} for FH{fitLength(FH, 3)} could not be retrieved.")
                sys.exit()
            except KeyboardInterrupt:
                raise KeyboardInterrupt()
            except:
                print('\a\nThe following error occurred:')
                execPacket = sys.exc_info()
                for i in execPacket:
                    print(i)

                print('\nPress ENTER to retry')
                print('Press Control-C to abort the program')
                input()

                continue

            print("\rProcessing member #" + str(member+1) + ", FH" + fitLength(FH, 3) + "           ", end="", flush=True)

            # Extract the variables we need.
            temperature = Field(modelData, "Temperature", "GFS", FH)
            humidity    = Field(modelData, "Relative humidity", "GFS", FH)
            omega       = Field(modelData, "Vertical velocity (geometric)", "GFS", FH)
            height      = Field(modelData, "Geopotential height", "GFS", FH)
            u           = Field(modelData, "u-component of wind", "GFS", FH)
            v           = Field(modelData, "v-component of wind", "GFS", FH)

            # Crop the grids so we don't have more data than we need.
            temperature.crop(subLatIndex, subLonIndex)
            humidity.crop(subLatIndex, subLonIndex)
            omega.crop(subLatIndex, subLonIndex)
            height.crop(subLatIndex, subLonIndex)
            u.crop(subLatIndex, subLonIndex)
            v.crop(subLatIndex, subLonIndex)

            temperatureBuffer.append(temperature)
            humidityBuffer.append(humidity)
            omegaBuffer.append(omega)
            heightBuffer.append(height)
            uBuffer.append(u)
            vBuffer.append(v)

            groundPressure = getSurfacePressure(height)

            if (pressureField == None):
                pressureField = height.level

                if (len(pressureField) == 0):
                    raise ValueError("No vertical profile")
        else:
            modelData = None

            try:
                modelData = loadModel("GEFS", year, month, day, hour, FH, memberID=member)
            except ValueError:
                print(f"\aERROR: Model data for the {fitLength(hour, 2)}Z run on {month}/{day}/{year} for FH{fitLength(FH, 3)} could not be retrieved.")
                sys.exit()
            except KeyboardInterrupt:
                raise KeyboardInterrupt()
            except:
                print('\a\nThe following error occurred:')
                execPacket = sys.exc_info()
                for i in execPacket:
                    print(i)

                print('\nPress ENTER to retry')
                print('Press Control-C to abort the program')
                input()

                continue

            print("\rProcessing member #" + str(member+1) + ", FH" + fitLength(FH, 3) + "   ", end="", flush=True)

            temperature = Field(modelData, "Temperature", "GEFS-" + str(member), FH)
            humidity    = Field(modelData, "Relative humidity", "GEFS-" + str(member), FH)

            # As it turns out, GEFS assumes hydrostatic balance, so there is no vertical
            # velocity data to retrieve.

            temperature.crop(subLatIndex, subLonIndex)
            humidity.crop(subLatIndex, subLonIndex)

            temperatureBuffer.append(temperature)
            humidityBuffer.append(humidity)

        member = member + 1

    # Maximum value for 'j' (a looping variable used below)
    maxJ = len(temperatureBuffer[0].data[0][0,:])
    N = len(temperatureBuffer[0].data[0][:,0])*len(temperatureBuffer[0].data[0][0,:])

    # Center these settings on the grid.
    offsetX = (len(temperatureBuffer[0].data[0][0,:]) % resolution) / 2.0
    offsetY = (len(temperatureBuffer[0].data[0][:,0]) % resolution) / 2.0

    if (int(offsetX) != offsetX):
        print('\aWARNING: x-dimension could not be centered')

    if (int(offsetY) != offsetY):
        print('\aWARNING: y-dimension could not be centered')

    # Initialize some empty arrays.
    probabilityGrid = np.zeros([len(omegaBuffer[0].data[0][:,0]), len(omegaBuffer[0].data[0][0,:])])
    windSpeed       = np.zeros([len(omegaBuffer[0].data[0][:,0]), len(omegaBuffer[0].data[0][0,:])])
    windDirection   = np.zeros([len(omegaBuffer[0].data[0][:,0]), len(omegaBuffer[0].data[0][0,:])])

    # Cycle through each latitude.
    for i in range(int(offsetY), len(temperatureBuffer[0].data[0][:,0]), resolution):
        # Cycle through each longitude.
        for j in range(int(offsetX), len(temperatureBuffer[0].data[0][0,:]), resolution):
            if (j + resolution - 1 >= len(temperatureBuffer[0].data[0][0,:])):
                continue

            if (i + resolution - 1 >= len(temperatureBuffer[0].data[0][:,0])):
                continue

            sigmaT = 0
            sigmaW = 0

            oldI = i
            oldJ = j

            soundingList = []

            NT = 0
            NW = 0

            # Query some forecast soundings near the center of the subsection we're
            # examining.
            for sY in range(i + (resolution // 2) - 1, i + (resolution // 2) + 1):
                for sX in range(j + (resolution // 2) - 1, j + (resolution // 2) + 1):
                    sounding = soundingGrid(pressureField, heightBuffer[0].getProfile(sY, sX), temperatureBuffer[0].getProfile(sY, sX), humidityBuffer[0].getProfile(sY, sX), uBuffer[0].getProfile(sY, sX), vBuffer[0].getProfile(sY, sX), omegaBuffer[0].getProfile(sY, sX))

                    # Go through each ensemble member to calculate standard deviation.
                    # NOTE: n = 1 is the ensemble mean.
                    for n in range(2, len(temperatureBuffer)):
                        sigmaT = sigmaT + (temperatureBuffer[n].data[0][sY, sX] - temperatureBuffer[1].data[0][sY, sX])**2
                        NT = NT + 1

                    # Calculate an uncertainty term for vertical velocity. NOTE: We cannot
                    # use the ensemble members for this since they assume hydrostatic
                    # balance.
                    sigmaW = sigmaW + omegaBuffer[0].getStandardDeviation(sY, sX)
                    NW = NW + 1

                    soundingList.append(sounding)

            # Take the soundings we called up and calculate a mean.
            soundingData = soundingMean(soundingList)

            sigmaT = math.sqrt(sigmaT / NT)         # Without dressing
            sigmaT = gain * sigmaT                  # With dressing

            sigmaW = math.sqrt(sigmaW / NW)         # Without dressing
            sigmaW = gain * sigmaW                  # With dressing

            # Progress display.
            print("\rCalculating analysis field... (FH" + str(FH) + ", " + str(i) + ", " + str(j) + "):  " + str(round(100 * ((i*maxJ + j*resolution) / N), 5)) + " %             ", end="", flush=True)

            # If a GRIB file is corrupted, data can be missing.
            missingData = np.where((np.isnan(soundingData)))
            soundingData[missingData] = -999

            w = 0
            wCount = 0

            # Go through this subsection and calculate the mean value of one-half the max 
            # vertical velocity.
            for sY in range(i, i + resolution):
                for sX in range(j, j + resolution):
                    w = w + 0.5 * omegaBuffer[0].getMax(sY, sX)
                    wCount = wCount + 1

            w = w / wCount
            T = soundingData[-1, 2]
            GLP = groundPressure[i:i+resolution, j:j+resolution].mean()

            # Now calculate this value of P0.
            P0 = getOverallProbability(GLP, soundingData, w, sigmaT, sigmaW)

            probabilityGrid[i:i+resolution, j:j+resolution] = P0

    thetaArray = thetaGrid(temperatureBuffer[0].data[0], humidityBuffer[0].data[0], pressureField[0]*100)

    frontAngle = frontAngleGrid(thetaArray, size=16)


    for i in range(0, len(temperatureBuffer[0].data[0][:,0])):
        for j in range(0, len(temperatureBuffer[0].data[0][0,:])):
            soundingData = soundingGrid(pressureField, heightBuffer[0].getProfile(i, j), temperatureBuffer[0].getProfile(i, j), humidityBuffer[0].getProfile(i, j), uBuffer[0].getProfile(i, j), vBuffer[0].getProfile(i, j), omegaBuffer[0].getProfile(i, j))

            meanSpeed, meanDirection = meanWind(soundingData, uBuffer[0].getProfile(i, j), vBuffer[0].getProfile(i, j))

            windSpeed[i, j] = meanSpeed
            windDirection[i, j] = meanDirection

    modeValue = np.zeros([len(windSpeed[:,0]), len(windSpeed[0,:])])

    modeValue = np.ma.array(modeValue, mask=list(modeValue))

    lowerLatitude = temperatureBuffer[0].latitude[-1, 0]
    upperLatitude = temperatureBuffer[0].latitude[0,  0]

    lowerLongitude = temperatureBuffer[0].longitude[0, 0]
    upperLongitude = temperatureBuffer[0].longitude[0, -1]

    latitude = np.arange(upperLatitude, lowerLatitude - 0.5, -0.5)
    longitude = np.arange(lowerLongitude, upperLongitude + 0.5, 0.5)

    month_day = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    dataHour = FH + settings['run-hour']
    dataDay = originalDay
    dataMonth = originalMonth
    dataYear = originalYear

    if (dataYear % 4) == 0:          # Leap year?
        month_day[2] = 29
    else:
        month_day[2] = 28

    if (dataDay > month_day[month]):
        dataDay = dataDay - month_day[month]
        dataMonth = dataMonth + 1

        if (dataMonth > 12):
            dataMonth = 1
            dataYear  = dataYear + 1

    while (dataHour >= 24):
        dataHour = dataHour - 24
        dataDay  = dataDay + 1

        if (dataDay > month_day[dataMonth]):
            dataDay   = 1
            dataMonth = dataMonth + 1

            if (dataMonth > 12):
                dataMonth = 1
                dataYear  = dataYear + 1

                if (dataYear % 4) == 0:
                    month_day[2] = 29
                else:
                    month_day[2] = 28


    timeStamp = fitLength(dataMonth, 2) + "-" + fitLength(dataDay, 2) + "-" + fitLength(dataYear, 4) + " " + fitLength(dataHour, 2) + "Z"

    for i in range(0, len(latitude)):
        for j in range(0, len(longitude)):
            front_theta = frontAngle[i, j]
            deltaY      = 139.99618625655287
            wind_theta  = windDirection[i, j]
            meanV       = windSpeed[i, j] * (1.15 / 0.621)
            storm_width = 35.84475806451613
            P0          = probabilityGrid[i, j]

            print("\rGenerating Forecast... (FH" + str(FH) + ", " + str(i) + ", " + str(j) + ")                         ", end="", flush=True)

            if (front_theta <= -999 or P0 == 0):
                A0 = (240)*(240)

                Ai = math.pi * (storm_width / 2)**2

                # Probability is too low to initiate any cells
                if (P0 < (0.5 * Ai) / A0):
                    modeValue.mask[i, j] = True

                    continue
                elif (P0 <= stochasticProbability[1]):
                    modeValue[i, j] = 0
                else:
                    xLinear = interpolate(stochastic, stochasticProbability, P0)

                    modeValue[i, j] = xLinear

            # There is a front here, so initialize the storms along the front.
            else:
                Y           = 256
                angle       = front_theta
                m           = tan(angle)

                if (m == 0):
                    m = 0.1

                X           = Y / m
                deltaX      = deltaY / m

                # Some geometry to estimate:
                # (1) The number of storms that will initiate along this front
                # (2) The initial distance between both storms
                #
                # Both of these quantities exclusively depend on P0 (all other values are
                # constant).

                A0          = Y * deltaY
                R           = 256
                r           = (math.pi * R * (storm_width / 2)**2) / (A0 * P0)
                N           = int(R / r) + 1

                # Probability is too low for any storms to form.
                if (round(R / r) <= 0):
                    modeValue.mask[i, j] = True

                    continue

                x       = []
                y       = []
                delta_s = []
                theta   = []
                v       = []

                for n in range(N):
                    x = [n * r * cos(angle)] + x
                    y = [n * r * sin(angle)] + y
                    delta_s.append(storm_width)
                    theta.append(wind_theta)
                    v.append(meanV)

                totalM = 0
                t_star = 0

                if (N >= 2):
                    # Calculate the time at which the storms are closest to each other.
                    t_star = ((x[0] - x[1]) * cos(theta[1])) / v[1] + ((y[0] - y[1]) * sin(theta[1])) / v[1]

                    # Cap 't_star' at 3 hours since we're only looking at 3-hour time
                    # windows.
                    if (t_star > 3.0):
                        t_star = 3.0
                    if (t_star < -3.0):
                        t_star = -3.0

                    # Now calculate the minimum spacing between the storms.
                    D_star = math.sqrt((x[1] + v[1] * t_star * cos(theta[1]) - x[0])**2 + (y[1] + v[1] * t_star * sin(theta[1]) - y[0])**2) - 0.5*delta_s[1] - 0.5*delta_s[0]

                    delta_s_bar = 0.5 * (delta_s[1] + delta_s[0])    # This is constant
                                                                     # since delta_s is
                                                                     # constant.

                    totalM = f(D_star, delta_s_bar)   # What does the graphical
                                                      # discriminant function say about
                                                      # storm mode?

                    modeValue[i, j] = totalM

                # Only have one storm, which is naturally 100% isolated.
                else:
                    modeValue[i, j] = 0


    if (settings['save-map']):
        directoryPathList = parentDirectory.split('/')
        directoryPrefix = directoryPathList[0] + '/'
        directoryPathList = directoryPathList[1:]

        while (not(os.path.exists(parentDirectory))):
            if (not(os.path.exists(directoryPrefix))):
                os.mkdir(directoryPrefix)

            directoryPrefix = directoryPrefix + directoryPathList[0] + '/'
            directoryPathList = directoryPathList[1:]

        # Create a customized colorbar.
        colors = [(1, 0, 0), (0.75, 0.75, 0.75), (0, 0, 1)]
        n_bins = [100]
        cmap_name = 'my_list'

        ax = plt.axes(projection = ccrs.Mercator())
        ax.set_title("CoMPS Storm Mode Forecast " + timeStamp)
        ax.coastlines()
        ax.set_global()
        ax.set_extent([lowerLongitude, upperLongitude, lowerLatitude, upperLatitude], crs=ccrs.PlateCarree())
        ax.add_feature(cf.COASTLINE)
        ax.add_feature(cf.STATES)

        for n_bin in n_bins:
            cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=n_bin)
            im = ax.imshow(modeValue, interpolation='nearest', origin='lower', cmap=cm, vmin = 0, vmax = 1)
            cp = ax.contourf(longitude, latitude, modeValue, transform=ccrs.PlateCarree(), cmap=cm, vmin = 0, vmax = 1, levels=100)
            cb = plt.colorbar(im, ax=ax, orientation="horizontal")  # Format and label the
            cb.set_ticks([0, 0.5, 1.0])                  # colorbar
            cb.set_ticklabels(["Isolated", "Mixed", "Linear"])

        plt.savefig(parentDirectory + stampPrefix + " Storm Mode Forecast "+timeStamp+".png")
        plt.clf()

    if (settings['save-grid']):
        directoryPathList = parentDirectory.split('/')
        directoryPrefix = directoryPathList[0] + '/'
        directoryPathList = directoryPathList[1:]

        while (not(os.path.exists(parentDirectory))):
            if (not(os.path.exists(directoryPrefix))):
                os.mkdir(directoryPrefix)

            directoryPrefix = directoryPrefix + directoryPathList[0] + '/'
            directoryPathList = directoryPathList[1:]

        # .cfg = CoMPS Forecast Grid
        makeDataset(longitude, latitude, modeValue, (parentDirectory + stampPrefix + "Storm Mode Forecast "+timeStamp+".cfg"))

    temperatureBuffer = []
    humidityBuffer = []
    omegaBuffer = []
    heightBuffer = []
    uBuffer = []
    vBuffer = []

    FH = FH + 3


print('\a\n\n [Process Complete] \n\n')
